# stdlib
from typing import Any

# third party
from packaging import version
import pytest
import torch

# syft absolute
import syft as sy
from syft.lib.torch.return_types import types_fields

A = torch.tensor([[1.0, 1, 1], [2, 3, 4], [3, 5, 2], [4, 2, 5], [5, 4, 3]])
B = torch.tensor([[-10.0, -3], [12, 14], [14, 12], [16, 16], [18, 16]])
x = torch.Tensor([[1, 2], [1, 2]])
s = torch.tensor(
    [[-0.1000, 0.1000, 0.2000], [0.2000, 0.3000, 0.4000], [0.0000, -0.3000, 0.5000]]
)

torch_version_ge_1d5d0 = version.parse(
    torch.__version__.split("+")[0]
) >= version.parse("1.5.0")

parameters = [
    ("eig", x, True),
    ("kthvalue", x, 1),
    ("lstsq", A, B),
    ("slogdet", x, None),
    ("qr", x, None),
    ("mode", x, None),
    ("solve", s, s),
    ("sort", s, None),
    ("symeig", s, None),
    ("topk", s, 1),
    ("triangular_solve", s, s),
    ("svd", s, None),
    ("geqrf", s, None),
    ("median", s, 0),
    ("max", s, 0),
    ("min", s, 0),
]

if torch_version_ge_1d5d0:
    parameters.append(("cummax", x, 0))
    parameters.append(("cummin", x, 0))


def assert_eq(y1, y2):  # type: ignore
    fields = types_fields[type(y1)]
    for field in fields:
        assert (getattr(y1, field) == getattr(y2, field)).all()


def assert_serde(y):  # type: ignore
    ser = sy.serialize(y)
    de = sy.deserialize(ser)
    assert_eq(y, de)


@pytest.mark.parametrize("op_name, ten, _arg", parameters)
def test_returntypes(
    op_name: str,
    ten: torch.Tensor,
    _arg: Any,
    node: sy.VirtualMachine,
    client: sy.VirtualMachineClient,
) -> None:
    # serde y=ten.op(_arg)
    op = getattr(ten, op_name)
    if _arg is not None:
        y = op(_arg)
    else:
        y = op()

    assert_serde(y)  # type: ignore

    # serde y=torch.Tensor.op(ten, _arg)
    op = getattr(torch.Tensor, op_name)
    if _arg is not None:
        y = op(ten, _arg)
    else:
        y = op(ten)

    assert_serde(y)  # type: ignore

    # ptr.op(_arg).get()
    ptr = ten.send(client)
    op = getattr(ptr, op_name)
    if _arg is not None:
        y_ptr = op(_arg)
    else:
        y_ptr = op()
    y_get = y_ptr.get()
    assert_eq(y, y_get)  # type: ignore

    # alice_client.torch.Tensor.op(ptr, _arg).get()
    op = getattr(client.torch.Tensor, op_name)
    if _arg is not None:
        y_ptr = op(ptr, _arg)
    else:
        y_ptr = op(ptr)
    y_get = y_ptr.get()
    assert_eq(y, y_get)  # type: ignore

    # TODO: Loot at https://github.com/OpenMined/PySyft/issues/5249
    # serde y=torch.op(ten, _arg)
    # alice_client.torch.op(ptr, _arg).get()



# stdlib
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Set as TypeSet

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ... import deserialize
from ... import serialize
from ...core.common.serde.serializable import bind_protobuf
from ...core.common.uid import UID
from ...proto.lib.python.set_pb2 import Set as Set_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet
from .util import downcast
from .util import upcast


@bind_protobuf
class Set(set, PyPrimitive):
    def __init__(self, iterable: Iterable, _id: Optional[UID] = None):
        super().__init__(iterable)

        self._id = UID() if _id is None else _id

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> TypeSet:
        # recursively upcast
        return {upcast(v) for v in self}

    def __and__(self, other: Any) -> SyPrimitiveRet:
        res = super().__and__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __contains__(self, other: Any) -> SyPrimitiveRet:
        res = super().__contains__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = super().__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ge__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__gt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __iand__(self, other: Any) -> SyPrimitiveRet:
        res = super().__iand__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ior__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ior__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __isub__(self, other: Any) -> SyPrimitiveRet:
        res = super().__isub__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ixor__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ixor__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        res = super().__le__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __len__(self) -> SyPrimitiveRet:
        res = super().__len__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__lt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __or__(self, other: Any) -> SyPrimitiveRet:
        res = super().__or__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __sub__(self, other: Any) -> SyPrimitiveRet:
        res = super().__sub__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __xor__(self, other: Any) -> SyPrimitiveRet:
        res = super().__xor__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def add(self, other: Any) -> None:
        res = super().add(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def clear(self) -> None:
        res = super().clear()
        return PrimitiveFactory.generate_primitive(value=res)

    def difference(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().difference(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def difference_update(self, *args: Any) -> None:
        res = super().difference_update(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def discard(self, element: Any) -> None:
        res = super().discard(element)
        return PrimitiveFactory.generate_primitive(value=res)

    def intersection(self, *args: Any) -> SyPrimitiveRet:
        res = super().intersection(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def intersection_update(self, *args: Any) -> None:
        res = super().intersection_update(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def isdisjoint(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().isdisjoint(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def issubset(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().issubset(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def issuperset(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().issuperset(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def pop(self) -> SyPrimitiveRet:
        res = super().pop()
        return PrimitiveFactory.generate_primitive(value=res)

    def remove(self, element: Any) -> None:
        res = super().remove(element)
        return PrimitiveFactory.generate_primitive(value=res)

    def symmetric_difference(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().symmetric_difference(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def symmetric_difference_update(self, s: Any) -> None:
        res = super().symmetric_difference_update(s)
        return PrimitiveFactory.generate_primitive(value=res)

    def union(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().union(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def update(self, *args: Any) -> None:
        res = super().update(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def _object2proto(self) -> Set_PB:
        id_ = serialize(obj=self.id)
        downcasted = [downcast(value=element) for element in self]
        data = [serialize(obj=element, to_bytes=True) for element in downcasted]
        return Set_PB(id=id_, data=data)

    @staticmethod
    def _proto2object(proto: Set_PB) -> "Set":
        id_: UID = deserialize(blob=proto.id)
        value = [
            upcast(deserialize(blob=element, from_bytes=True)) for element in proto.data
        ]
        new_list = Set(value)
        new_list._id = id_
        return new_list

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Set_PB



# third party
import pytest
import torch as th

# syft absolute
import syft as sy
from syft import Plan
from syft import PlanBuilder
from syft import make_plan
from syft import serialize
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.node.common.action.common import Action
from syft.core.node.common.action.function_or_constructor_action import (
    RunFunctionOrConstructorAction,
)
from syft.core.node.common.action.garbage_collect_object_action import (
    GarbageCollectObjectAction,
)
from syft.core.node.common.action.get_enum_attribute_action import EnumAttributeAction
from syft.core.node.common.action.get_object_action import GetObjectAction
from syft.core.node.common.action.get_or_set_property_action import (
    GetOrSetPropertyAction,
)
from syft.core.node.common.action.get_or_set_property_action import PropertyActions
from syft.core.node.common.action.get_or_set_static_attribute_action import (
    GetSetStaticAttributeAction,
)
from syft.core.node.common.action.get_or_set_static_attribute_action import (
    StaticAttributeAction,
)
from syft.core.node.common.action.run_class_method_action import RunClassMethodAction
from syft.core.node.common.action.save_object_action import SaveObjectAction
from syft.core.plan.plan_builder import ROOT_CLIENT
from syft.core.store.storeable_object import StorableObject
from syft.lib.python.list import List


def test_plan_serialization(client: sy.VirtualMachineClient) -> None:

    # cumbersome way to get a pointer as input for our actions,
    # there is probably a better/shorter way
    t = th.tensor([1, 2, 3])
    tensor_pointer = t.send(client)

    # define actions
    a1 = GetObjectAction(
        id_at_location=UID(), address=Address(), reply_to=Address(), msg_id=UID()
    )
    a2 = RunFunctionOrConstructorAction(
        path="torch.Tensor.add",
        args=tuple(),
        kwargs={},
        id_at_location=UID(),
        address=Address(),
        msg_id=UID(),
    )

    a3 = RunClassMethodAction(
        path="torch.Tensor.add",
        _self=tensor_pointer,
        args=[],
        kwargs={},
        id_at_location=UID(),
        address=Address(),
        msg_id=UID(),
    )

    a4 = GarbageCollectObjectAction(id_at_location=UID(), address=Address())
    a5 = EnumAttributeAction(path="", id_at_location=UID(), address=Address())

    a6 = GetOrSetPropertyAction(
        path="",
        _self=tensor_pointer,
        id_at_location=UID(),
        address=Address(),
        args=[],
        kwargs={},
        action=PropertyActions.GET,
    )
    a7 = GetSetStaticAttributeAction(
        path="",
        id_at_location=UID(),
        address=Address(),
        action=StaticAttributeAction.GET,
    )
    a8 = SaveObjectAction(obj=StorableObject(id=UID(), data=t), address=Address())

    # define plan
    plan = Plan([a1, a2, a3, a4, a5, a6, a7, a8])

    # serialize / deserialize
    blob = serialize(plan)
    plan_reconstructed = sy.deserialize(blob=blob)

    # test
    assert isinstance(plan_reconstructed, Plan)
    assert all(isinstance(a, Action) for a in plan_reconstructed.actions)


def test_plan_execution(client: sy.VirtualMachineClient) -> None:
    tensor_pointer1 = th.tensor([1, 2, 3]).send(client)
    tensor_pointer2 = th.tensor([4, 5, 6]).send(client)
    tensor_pointer3 = th.tensor([7, 8, 9]).send(client)

    result_tensor_pointer1 = th.tensor([0, 0, 0]).send(client)
    result_tensor_pointer2 = th.tensor([0, 0, 0]).send(client)

    result1_uid = result_tensor_pointer1.id_at_location
    result2_uid = result_tensor_pointer2.id_at_location

    a1 = RunClassMethodAction(
        path="torch.Tensor.add",
        _self=tensor_pointer1,
        args=[tensor_pointer2],
        kwargs={},
        id_at_location=result1_uid,
        address=Address(),
        msg_id=UID(),
    )

    a2 = RunClassMethodAction(
        path="torch.Tensor.add",
        _self=result_tensor_pointer1,
        args=[tensor_pointer3],
        kwargs={},
        id_at_location=result2_uid,
        address=Address(),
        msg_id=UID(),
    )

    plan = Plan([a1, a2])

    plan_pointer = plan.send(client)

    plan_pointer()

    expected_tensor1 = th.tensor([5, 7, 9])
    expected_tensor2 = th.tensor([12, 15, 18])

    assert all(expected_tensor1 == result_tensor_pointer1.get())
    assert all(expected_tensor2 == result_tensor_pointer2.get())


def test_plan_batched_execution(client: sy.VirtualMachineClient) -> None:
    # placeholders for our input
    input_tensor_pointer1 = th.tensor([0, 0]).send(client)
    input_tensor_pointer2 = th.tensor([0, 0]).send(client)

    # tensors in our model
    model_tensor_pointer1 = th.tensor([1, 2]).send(client)
    model_tensor_pointer2 = th.tensor([3, 4]).send(client)

    # placeholders for intermediate results
    result_tensor_pointer1 = th.tensor([0, 0]).send(client)
    result_tensor_pointer2 = th.tensor([0, 0]).send(client)
    result_tensor_pointer3 = th.tensor([0, 0]).send(client)

    # define plan
    a1 = RunClassMethodAction(
        path="torch.Tensor.mul",
        _self=input_tensor_pointer1,
        args=[model_tensor_pointer1],
        kwargs={},
        id_at_location=result_tensor_pointer1.id_at_location,
        address=Address(),
        msg_id=UID(),
    )

    a2 = RunClassMethodAction(
        path="torch.Tensor.add",
        _self=result_tensor_pointer1,
        args=[model_tensor_pointer2],
        kwargs={},
        id_at_location=result_tensor_pointer2.id_at_location,
        address=Address(),
        msg_id=UID(),
    )

    a3 = RunFunctionOrConstructorAction(
        path="torch.eq",
        args=[result_tensor_pointer2, input_tensor_pointer2],
        kwargs={},
        id_at_location=result_tensor_pointer3.id_at_location,
        address=Address(),
        msg_id=UID(),
    )

    plan = Plan(
        [a1, a2, a3], inputs={"x": input_tensor_pointer1, "y": input_tensor_pointer2}
    )
    plan_pointer = plan.send(client)

    # Test
    # x is random input, y is the expected model(x)
    x_batches = [(th.tensor([1, 1]) + i).send(client) for i in range(2)]
    y_batches = [
        ((th.tensor([1, 1]) + i) * th.tensor([1, 2]) + th.tensor([3, 4])).send(client)
        for i in range(2)
    ]

    for x, y in zip(x_batches, y_batches):
        plan_pointer(x=x, y=y)

        # checks if (model(x) == y) == [True, True]
        assert all(result_tensor_pointer3.get(delete_obj=False))


def test_make_plan(client: sy.VirtualMachineClient) -> None:
    @make_plan
    def add_plan(inp=th.zeros((3))) -> th.Tensor:  # type: ignore
        return inp + inp

    input_tensor = th.tensor([1, 2, 3])
    plan_pointer = add_plan.send(client)
    res = plan_pointer(inp=input_tensor)
    assert th.equal(res.get()[0], th.tensor([2, 4, 6]))


@pytest.mark.xfail
def test_plan_deterministic_bytes(root_client: sy.VirtualMachineClient) -> None:
    # TODO: https://github.com/OpenMined/PySyft/issues/5292
    @make_plan
    def add_plan(inp=th.zeros((3))) -> th.Tensor:  # type: ignore
        return inp + inp

    @make_plan
    def add_plan2(inp=th.zeros((3))) -> th.Tensor:  # type: ignore
        return inp + inp

    plan_pointer = add_plan.send(root_client)
    plan2_pointer = add_plan2.send(root_client)

    plan1 = serialize(plan_pointer, to_bytes=True)
    plan2 = serialize(plan2_pointer, to_bytes=True)

    assert plan1 == plan2


def test_make_plan2(root_client: sy.VirtualMachineClient) -> None:
    @make_plan
    def mul_plan(inp=th.zeros((3)), inp2=th.zeros((3))) -> th.Tensor:  # type: ignore
        return inp * inp2

    t1, t2 = th.tensor([1, 2, 3]), th.tensor([1, 2, 3])
    plan_pointer = mul_plan.send(root_client)
    res = plan_pointer(inp=t1, inp2=t2)
    assert th.equal(res.get()[0], th.tensor([1, 4, 9]))


def test_make_plan_error_no_kwargs() -> None:
    def assertRaises(exc, obj, methodname, *args) -> None:  # type: ignore
        with pytest.raises(exc) as e_info:
            getattr(obj, methodname)(*args)
        assert str(e_info) != ""

    def test_define_plan():  # type: ignore
        @make_plan
        def add_plan(inp):  # type: ignore
            return inp + inp

    # we can only define a plan with *kwargs* when using the @make_plan decorator, not with args
    assertRaises(ValueError, test_define_plan, "__call__")


def test_mlp_plan(client: sy.VirtualMachineClient) -> None:
    class MLP(sy.Module):
        def __init__(self, torch_ref):  # type: ignore
            super().__init__(torch_ref=torch_ref)
            self.l1 = self.torch_ref.nn.Linear(784, 100)
            self.a1 = self.torch_ref.nn.ReLU()
            self.l2 = self.torch_ref.nn.Linear(100, 10)

        def forward(self, x):  # type: ignore
            x_reshaped = x.view(-1, 28 * 28)
            l1_out = self.a1(self.l1(x_reshaped))
            l2_out = self.l2(l1_out)
            return l2_out

    def set_params(model, params):  # type: ignore
        """happens outside of plan"""
        for p, p_new in zip(model.parameters(), params):
            p.data = p_new.data

    def cross_entropy_loss(logits, targets, batch_size):  # type: ignore
        norm_logits = logits - logits.max()
        log_probs = norm_logits - norm_logits.exp().sum(dim=1, keepdim=True).log()
        return -(targets * log_probs).sum() / batch_size

    def sgd_step(model, lr=0.1):  # type: ignore
        with ROOT_CLIENT.torch.no_grad():
            for p in model.parameters():
                p.data = p.data - lr * p.grad
                # Todo: fix this
                p.grad = th.zeros_like(p.grad.get())

    local_model = MLP(th)  # type: ignore

    @make_plan
    def train(  # type: ignore
        xs=th.rand([64 * 3, 1, 28, 28]),
        ys=th.randint(0, 10, [64 * 3, 10]),
        params=List(local_model.parameters()),
    ):

        model = local_model.send(ROOT_CLIENT)
        set_params(model, params)
        for i in range(2):
            indices = th.tensor(range(64 * i, 64 * (i + 1)))
            x, y = xs.index_select(0, indices), ys.index_select(0, indices)
            out = model(x)
            loss = cross_entropy_loss(out, y, 64)
            loss.backward()
            sgd_step(model)

        return model.parameters()

    train_ptr = train.send(client)

    old_params = local_model.parameters().copy()

    res_ptr = train_ptr(
        xs=th.rand([64 * 3, 1, 28, 28]),
        ys=th.randint(0, 10, [64 * 3, 10]),
        params=local_model.parameters(),
    )

    (new_params,) = res_ptr.get()

    assert not (old_params[0] == new_params[0]).all()


def test_planbuilder() -> None:
    class TestModel(PlanBuilder):
        def __init__(self) -> None:
            self.model_pointer = th.tensor([1, 2, 3])
            super().__init__()

        def forward(self, x: th.Tensor = th.tensor([0, 0, 0])) -> th.Tensor:
            res = x * self.model_pointer
            return res

    model = TestModel()

    res = model(x=th.tensor([4, 5, 6]))
    assert th.equal(*res.get(), th.tensor([4, 10, 18]))



# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pytest

# syft absolute
import syft as sy
from syft.core.node.common.service.auth import AuthorizationException


def get_signing_key() -> SigningKey:
    # return a the signing key to use
    key = "e89ff2e651b42393b6ecb5956419088781309d953d72bd73a0968525a3a6a951"
    return SigningKey(bytes.fromhex(key))


def get_verify_key() -> VerifyKey:
    return get_signing_key().verify_key


def test_to_string(node: sy.VirtualMachine) -> None:
    assert str(node) == f"VirtualMachine: Bob: {node.id}"
    assert node.__repr__() == f"VirtualMachine: Bob: {node.id}"


def test_send_message_from_vm_client_to_vm(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    assert node.device is None

    with pytest.raises(AuthorizationException):
        client.send_immediate_msg_without_reply(
            msg=sy.ReprMessage(address=client.address)
        )


def test_send_message_from_device_client_to_device() -> None:
    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()

    with pytest.raises(AuthorizationException):
        bob_phone_client.send_immediate_msg_without_reply(
            msg=sy.ReprMessage(address=bob_phone_client.address)
        )


@pytest.mark.slow
def test_register_vm_on_device_fails(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()

    with pytest.raises(AuthorizationException):
        bob_phone_client.register(client=client)

    assert node.device is None

    # TODO: prevent device being set when Authorization fails
    assert client.device is not None


@pytest.mark.slow
def test_register_vm_on_device_succeeds() -> None:
    # Register a 🍰 with a 📱

    bob_vm = sy.VirtualMachine(name="Bob")
    bob_vm_client = bob_vm.get_client()
    bob_vm.root_verify_key = bob_vm_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()
    bob_phone.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.register(client=bob_vm_client)

    assert bob_vm.device is not None
    assert bob_vm_client.device is not None


@pytest.mark.slow
def test_known_child_nodes() -> None:
    bob_vm = sy.VirtualMachine(name="Bob VM")
    bob_vm_client = bob_vm.get_client()
    bob_vm.root_verify_key = bob_vm_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_vm_2 = sy.VirtualMachine(name="Bob VM 2")
    bob_vm_client_2 = bob_vm_2.get_client()
    bob_vm_2.root_verify_key = bob_vm_client_2.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()
    bob_phone.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.register(client=bob_vm_client)

    assert len(bob_phone.known_child_nodes) == 1
    assert bob_vm in bob_phone.known_child_nodes

    bob_phone_client.register(client=bob_vm_client_2)

    assert len(bob_phone.known_child_nodes) == 2
    assert bob_vm_2 in bob_phone.known_child_nodes


@pytest.mark.slow
def test_send_message_from_device_client_to_vm() -> None:
    # Register a 🍰 with a 📱
    # Send ✉️ from 📱 ➡️ 🍰

    bob_vm = sy.VirtualMachine(name="Bob")
    bob_vm_client = bob_vm.get_client()
    bob_vm.root_verify_key = bob_vm_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()
    bob_phone.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.register(client=bob_vm_client)

    assert bob_vm.device is not None
    assert bob_vm_client.device is not None

    # switch keys
    bob_vm.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.send_immediate_msg_without_reply(
        msg=sy.ReprMessage(address=bob_vm_client.address)
    )


@pytest.mark.slow
@pytest.mark.asyncio
def test_send_message_from_domain_client_to_vm() -> None:
    # Register a 🍰 with a 📱
    # Register a 📱 with a 🏰
    # Send ✉️ from 🏰 ➡️ 🍰

    bob_vm = sy.VirtualMachine(name="Bob")
    bob_vm_client = bob_vm.get_client()
    bob_vm.root_verify_key = bob_vm_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()
    bob_phone.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.register(client=bob_vm_client)

    assert bob_vm.device is not None
    assert bob_vm_client.device is not None

    bob_domain = sy.Domain(name="Bob's Domain")
    bob_domain_client = bob_domain.get_client()
    bob_domain.root_verify_key = bob_domain_client.verify_key  # inject 📡🔑 as 📍🗝

    # switch keys
    bob_vm.root_verify_key = bob_domain_client.verify_key  # inject 📡🔑 as 📍🗝
    bob_domain_client.register(client=bob_phone_client)

    assert bob_phone.domain is not None
    assert bob_phone_client.domain is not None

    bob_domain_client.send_immediate_msg_without_reply(
        msg=sy.ReprMessage(address=bob_vm.address)
    )


@pytest.mark.slow
@pytest.mark.asyncio
def test_send_message_from_network_client_to_vm() -> None:
    # Register a 🍰 with a 📱
    # Register a 📱 with a 🏰
    # Register a 🏰 with a 🔗
    # Send ✉️ from 🔗 ➡️ 🍰

    bob_vm = sy.VirtualMachine(name="Bob")
    bob_vm_client = bob_vm.get_client()
    bob_vm.root_verify_key = bob_vm_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()
    bob_phone.root_verify_key = bob_phone_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_phone_client.register(client=bob_vm_client)

    assert bob_vm.device is not None
    assert bob_vm_client.device is not None

    bob_domain = sy.Domain(name="Bob's Domain")
    bob_domain_client = bob_domain.get_client()
    bob_domain.root_verify_key = bob_domain_client.verify_key  # inject 📡🔑 as 📍🗝

    bob_domain_client.register(client=bob_phone_client)

    assert bob_phone.domain is not None
    assert bob_phone_client.domain is not None

    bob_network = sy.Network(name="Bob's Network")
    bob_network_client = bob_network.get_client()
    bob_network.root_verify_key = bob_network_client.verify_key  # inject 📡🔑 as 📍🗝

    # switch keys
    bob_vm.root_verify_key = bob_network_client.verify_key  # inject 📡🔑 as 📍🗝
    bob_network_client.register(client=bob_domain_client)

    assert bob_domain.network is not None
    assert bob_domain_client.network is not None

    bob_network_client.send_immediate_msg_without_reply(
        msg=sy.ReprMessage(address=bob_vm.address)
    )


@pytest.mark.slow
@pytest.mark.asyncio
def test_autoapprove_requests_made_by_root_clients_5015(
    client: sy.VirtualMachineClient, root_client: sy.VirtualMachineClient
) -> None:
    # third party
    import torch

    p = root_client.torch.Tensor([1, 2, 3])
    t = p.get(request_block=True, reason="Test")
    assert torch.equal(t, torch.Tensor([1, 2, 3]))

    p = client.torch.Tensor([1, 2, 3])
    t = p.get(request_block=True, reason="Test")
    assert t is None



# stdlib
from typing import List

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from syft_proto.execution.v1.state_pb2 import State as StatePB
from syft_proto.execution.v1.state_tensor_pb2 import StateTensor as StateTensorPB

# syft relative
from ...core.common.object import Serializable
from ...core.common.serde.serialize import _serialize as serialize
from .common import deserialize_tensor
from .common import serialize_tensor
from .placeholder import PlaceHolder


class State(Serializable):
    """The State is a Plan attribute and is used to send tensors along functions.

    It references Plan tensor or parameters attributes using their name, and make
    sure they are provided to remote workers who are sent the Plan.
    """

    def __init__(self, state_placeholders: List[PlaceHolder] = []) -> None:
        self.state_placeholders = state_placeholders

    def tensors(self) -> List:
        """
        Fetch and return all the state elements.
        """
        return [placeholder.child for placeholder in self.state_placeholders]

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """
        return StatePB

    def _object2proto(self) -> StatePB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectWithID_PB

        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        proto = StatePB()
        protobuf_placeholders = [
            serialize(placeholder) for placeholder in self.state_placeholders
        ]
        proto.placeholders.extend(protobuf_placeholders)

        state_tensors = []
        for tensor in self.tensors():
            state_tensor = StateTensorPB()
            state_tensor.torch_tensor.CopyFrom(serialize_tensor(tensor))
            state_tensors.append(state_tensor)

        proto.tensors.extend(state_tensors)
        return proto

    @staticmethod
    def _proto2object(proto: StatePB) -> "State":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of Plan
        :rtype: Plan

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        state_placeholders = proto.placeholders
        state_elements = proto.tensors

        state_placeholders = [
            PlaceHolder._proto2object(placeholder) for placeholder in proto.placeholders
        ]

        state_elements = []
        for protobuf_tensor in proto.tensors:
            tensor = getattr(protobuf_tensor, protobuf_tensor.WhichOneof("tensor"))
            state_elements.append(deserialize_tensor(tensor))

        for state_placeholder, state_element in zip(state_placeholders, state_elements):
            state_placeholder.instantiate(state_element)

        state = State(state_placeholders)
        return state



# third party
import pytest
from pytest import approx

# syft absolute
import syft as sy


@pytest.mark.parametrize("loadlib_before_client", [True, False])
@pytest.mark.parametrize("reveal_intersection", [True, False])
@pytest.mark.vendor(lib="openmined_psi")
def test_psi(
    loadlib_before_client: bool, reveal_intersection: bool, node: sy.VirtualMachine
) -> None:
    # third party
    import openmined_psi as psi

    # it should work when call load before or after create clients
    if loadlib_before_client:
        sy.load("openmined_psi")
        server_vm = node.get_root_client()
        client_vm = node.get_root_client()
    else:
        server_vm = node.get_root_client()
        client_vm = node.get_root_client()
        sy.load("openmined_psi")

    # server send reveal_intersection
    s_reveal_intersection = reveal_intersection
    s_sy_reveal_intersection = sy.lib.python.Bool(s_reveal_intersection)
    s_sy_reveal_intersection.send(
        server_vm,
        pointable=True,
        tags=["reveal_intersection"],
        description="reveal intersection value",
    )
    assert (
        server_vm.store["reveal_intersection"].description
        == "reveal intersection value"
    )

    # client get reval_intersection
    c_reveal_intersection = server_vm.store["reveal_intersection"].get()
    assert c_reveal_intersection == s_reveal_intersection

    # server send fpr
    s_fpr = 1e-6
    s_sy_fpr = sy.lib.python.Float(s_fpr)
    s_sy_fpr.send(
        server_vm, pointable=True, tags=["fpr"], description="false positive rate"
    )

    # client get fpr
    c_fpr = server_vm.store["fpr"].get()
    assert c_fpr == approx(s_fpr)

    # client send client_items_len
    psi_client = psi.client.CreateWithNewKey(c_reveal_intersection)
    c_items = ["Element " + str(i) for i in range(1000)]
    c_sy_client_items_len = sy.lib.python.Int(len(c_items))
    c_sy_client_items_len.send(
        client_vm,
        pointable=True,
        tags=["client_items_len"],
        description="client items length",
    )

    # server get client_items_len
    s_sy_client_items_len = client_vm.store["client_items_len"].get(delete_obj=False)
    assert s_sy_client_items_len == c_sy_client_items_len

    # server send setup message
    s_items = ["Element " + str(2 * i) for i in range(1000)]
    psi_server = psi.server.CreateWithNewKey(s_reveal_intersection)
    s_setup = psi_server.CreateSetupMessage(s_fpr, s_sy_client_items_len, s_items)
    s_setup.send(
        server_vm,
        pointable=True,
        tags=["setup"],
        description="psi.server Setup Message",
    )
    assert server_vm.store["setup"].description == "psi.server Setup Message"

    # client get setup message
    c_setup = server_vm.store["setup"].get()
    assert c_setup == s_setup

    # client send request
    c_request = psi_client.CreateRequest(c_items)
    c_request.send(
        client_vm, tags=["request"], pointable=True, description="client request"
    )

    # server get request
    s_request = client_vm.store["request"].get()
    assert s_request == c_request

    # server send response
    s_response = psi_server.ProcessRequest(s_request)
    s_response.send(
        server_vm, pointable=True, tags=["response"], description="psi.server response"
    )

    # client get response
    c_response = server_vm.store["response"].get()
    assert c_response == s_response

    # client get result
    if c_reveal_intersection:
        intersection = psi_client.GetIntersection(c_setup, c_response)
        iset = set(intersection)
        for idx in range(len(c_items)):
            if idx % 2 == 0:
                assert idx in iset
            else:
                assert idx not in iset
    else:
        intersection = psi_client.GetIntersectionSize(c_setup, c_response)
        assert intersection >= (len(c_items) / 2.0)
        assert intersection <= (1.1 * len(c_items) / 2.0)



# stdlib
import platform
import re
import sys
from typing import Any
from typing import Dict

# third party
from packaging import version

VERSIONS_NONE: Dict[str, Any] = dict(torchvision=None, torchcsprng=None)
VERSIONS_LUT: Dict[str, Dict[str, Any]] = {
    "1.4.0": dict(torchvision="0.5.0", torchcsprng=None),
    "1.5.0": dict(torchvision="0.6.0", torchcsprng=None),
    "1.5.1": dict(torchvision="0.6.1", torchcsprng=None),
    "1.6.0": dict(torchvision="0.7", torchcsprng="0.1.2"),
    "1.7.0": dict(torchvision="0.8.1", torchcsprng="0.1.3"),
    "1.7.1": dict(torchvision="0.8.2", torchcsprng="0.1.4"),
    "1.8.0": dict(torchvision="0.9.0", torchcsprng="0.2.0"),
}

system = platform.system()


def main(path_req: str, torch_version: str) -> None:
    with open(path_req, "r") as fp:
        req = fp.read()

    dep_versions = VERSIONS_LUT.get(torch_version, VERSIONS_NONE)
    dep_versions["torch"] = torch_version
    for lib in dep_versions:
        replace = ""
        if dep_versions[lib]:
            dep_version = dep_versions[lib]
            lib_version = version.parse(dep_version)
            if system != "Darwin":
                dep_version += "+cpu"  # linux and windows require +cpu
            replace = f"{lib}=={dep_version}\n"

        if (
            lib == "torchcsprng"
            and sys.version_info >= (3, 9)
            and lib_version < version.parse("0.2.0")
        ):
            replace = ""  # no torchcsprng < 0.2.0 for python 3.9

        req = re.sub(
            rf"{lib}[<>=].*\n",
            replace,
            req,
        )

    with open(path_req, "w") as fp:
        fp.write(req)


if __name__ == "__main__":
    main(*sys.argv[1:])



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
from packaging import version
import pandas as pd

# syft relative
from . import frame  # noqa: 401
from . import series  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..misc.union import UnionGenerator
from ..util import generic_update_ast

LIB_NAME = "pandas"
PACKAGE_SUPPORT = {"lib": LIB_NAME}

LIB_VERSION = version.parse(pd.__version__.split("+")[0])


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [("pandas", pd)]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("pandas.DataFrame", "pandas.DataFrame", pd.DataFrame),
        ("pandas.Series", "pandas.Series", pd.Series),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        ("pandas.read_csv", "pandas.DataFrame"),
        ("pandas.DataFrame.__getitem__", "pandas.Series"),
        ("pandas.DataFrame.__setitem__", "pandas.Series"),
        ("pandas.DataFrame.__len__", "syft.lib.python.Int"),
        ("pandas.DataFrame.__abs__", "pandas.DataFrame"),
        ("pandas.DataFrame.__add__", "pandas.DataFrame"),
        ("pandas.DataFrame.__and__", "pandas.DataFrame"),
        ("pandas.DataFrame.__eq__", "pandas.DataFrame"),
        ("pandas.DataFrame.__floordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ge__", "pandas.DataFrame"),
        ("pandas.DataFrame.__gt__", "pandas.DataFrame"),
        ("pandas.DataFrame.__iadd__", "pandas.DataFrame"),
        ("pandas.DataFrame.__iand__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ifloordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__imod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__imul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ipow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__isub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__le__", "pandas.DataFrame"),
        ("pandas.DataFrame.__lt__", "pandas.DataFrame"),
        ("pandas.DataFrame.__mod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__mul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ne__", "pandas.DataFrame"),
        ("pandas.DataFrame.__neg__", "pandas.DataFrame"),
        ("pandas.DataFrame.__pos__", "pandas.DataFrame"),
        ("pandas.DataFrame.__pow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rfloordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rmod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rmul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__round__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rpow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rsub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rtruediv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__sub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__truediv__", "pandas.DataFrame"),
        ("pandas.DataFrame.dropna", "pandas.DataFrame"),
        ("pandas.Series.__getitem__", "pandas.Series"),
        ("pandas.Series.__setitem__", "pandas.Series"),
        ("pandas.Series.__len__", "syft.lib.python.Int"),
        ("pandas.Series.__abs__", "pandas.Series"),
        ("pandas.Series.__add__", "pandas.Series"),
        ("pandas.Series.__and__", "pandas.Series"),
        ("pandas.Series.__divmod__", "pandas.Series"),
        ("pandas.Series.__eq__", "pandas.Series"),
        ("pandas.Series.__floordiv__", "pandas.Series"),
        ("pandas.Series.__ge__", "pandas.Series"),
        ("pandas.Series.__gt__", "pandas.Series"),
        ("pandas.Series.__iadd__", "pandas.Series"),
        ("pandas.Series.__iand__", "pandas.Series"),
        ("pandas.Series.__ifloordiv__", "pandas.Series"),
        ("pandas.Series.__imod__", "pandas.Series"),
        ("pandas.Series.__imul__", "pandas.Series"),
        ("pandas.Series.__ipow__", "pandas.Series"),
        ("pandas.Series.__isub__", "pandas.Series"),
        ("pandas.Series.__le__", "pandas.Series"),
        ("pandas.Series.__lt__", "pandas.Series"),
        ("pandas.Series.__mod__", "pandas.Series"),
        ("pandas.Series.__mul__", "pandas.Series"),
        ("pandas.Series.__ne__", "pandas.Series"),
        ("pandas.Series.__neg__", "pandas.Series"),
        ("pandas.Series.__pos__", "pandas.Series"),
        ("pandas.Series.__pow__", "pandas.Series"),
        ("pandas.Series.__rdivmod__", "pandas.Series"),
        ("pandas.Series.__rfloordiv__", "pandas.Series"),
        ("pandas.Series.__rmod__", "pandas.Series"),
        ("pandas.Series.__rmul__", "pandas.Series"),
        ("pandas.Series.__round__", "pandas.Series"),
        ("pandas.Series.__rpow__", "pandas.Series"),
        ("pandas.Series.__rsub__", "pandas.Series"),
        ("pandas.Series.__rtruediv__", "pandas.Series"),
        ("pandas.Series.__sub__", "pandas.Series"),
        ("pandas.Series.__truediv__", "pandas.Series"),
        ("pandas.Series.add", "pandas.Series"),
        ("pandas.Series.sub", "pandas.Series"),
        ("pandas.Series.mul", "pandas.Series"),
        ("pandas.Series.div", "pandas.Series"),
        ("pandas.Series.truediv", "pandas.Series"),
        ("pandas.Series.floordiv", "pandas.Series"),
        ("pandas.Series.mod", "pandas.Series"),
        ("pandas.Series.pow", "pandas.Series"),
        ("pandas.Series.radd", "pandas.Series"),
        ("pandas.Series.rsub", "pandas.Series"),
        ("pandas.Series.rmul", "pandas.Series"),
        ("pandas.Series.rdiv", "pandas.Series"),
        ("pandas.Series.rtruediv", "pandas.Series"),
        ("pandas.Series.rfloordiv", "pandas.Series"),
        ("pandas.Series.rmod", "pandas.Series"),
        ("pandas.Series.rpow", "pandas.Series"),
        ("pandas.Series.lt", "pandas.Series"),
        ("pandas.Series.gt", "pandas.Series"),
        ("pandas.Series.le", "pandas.Series"),
        ("pandas.Series.ge", "pandas.Series"),
        ("pandas.Series.ne", "pandas.Series"),
        ("pandas.Series.eq", "pandas.Series"),
        ("pandas.Series.argsort", "pandas.Series"),
        ("pandas.Series.round", "pandas.Series"),
        ("pandas.Series.head", "pandas.Series"),
        ("pandas.Series.tail", "pandas.Series"),
        ("pandas.Series.any", "syft.lib.python.Bool"),
        ("pandas.Series.shape", "syft.lib.python.Tuple"),
        ("pandas.Series.all", "syft.lib.python.Bool"),
        ("pandas.Series.argmax", "syft.lib.python.Int"),
        ("pandas.Series.nbytes", "syft.lib.python.Int"),
        ("pandas.Series.mean", "syft.lib.python.Float"),
        ("pandas.Series.ndim", "syft.lib.python.Int"),
        ("pandas.Series.size", "syft.lib.python.Int"),
        ("pandas.Series.hasnans", "syft.lib.python.Bool"),
        ("pandas.Series.empty", "syft.lib.python.Bool"),
        ("pandas.Series.T", "pandas.Series"),
        ("pandas.Series.dropna", "pandas.Series"),
        ("pandas.Series.to_frame", "pandas.DataFrame"),
        ("pandas.Series.to_list", "syft.lib.python.List"),
        (
            "pandas.Series.sum",
            UnionGenerator["syft.lib.python.Float", "syft.lib.python.Int"],
        ),
        ("pandas.Series.median", "syft.lib.python.Float"),
        (
            "pandas.Series.max",
            UnionGenerator[
                "syft.lib.python.Bool", "syft.lib.python.Float", "syft.lib.python.Int"
            ],
        ),
        (
            "pandas.Series.min",
            UnionGenerator[
                "syft.lib.python.Bool", "syft.lib.python.Float", "syft.lib.python.Int"
            ],
        ),
    ]

    if LIB_VERSION > version.parse("1.2.0"):
        methods += [
            ("pandas.DataFrame.__divmod__", "pandas.DataFrame"),
            ("pandas.DataFrame.__rdivmod__", "pandas.DataFrame"),
        ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        # TODO: Pandas can't have tags and description because they break the dict
        # klass.create_storable_object_attr_convenience_methods()

    return ast


# we cant create Unions that refer to the package itself until the create_ast
# has completed first so we can call again into post_update_ast to finish these
# TODO: add support for self referential unions using some kind of post update
# Issue: https://github.com/OpenMined/PySyft/issues/5323
def post_create_ast(ast: Globals) -> Globals:
    self_referencing_methods = [
        ("pandas.Series.loc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.Series.iloc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.DataFrame.loc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.DataFrame.iloc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
    ]

    add_methods(ast, self_referencing_methods)

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)
# post_update_ast = functools.partial(generic_update_ast, LIB_NAME, post_create_ast)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/service/object_search_permission_update_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/service/object_search_permission_update_message.proto",
    package="syft.core.node.common.service",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\nLproto/core/node/common/service/object_search_permission_update_message.proto\x12\x1dsyft.core.node.common.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xdf\x01\n#ObjectSearchPermissionUpdateMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x19\n\x11target_verify_key\x18\x03 \x01(\x0c\x12/\n\x10target_object_id\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x1d\n\x15\x61\x64\x64_instead_of_remove\x18\x05 \x01(\x08\x62\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_OBJECTSEARCHPERMISSIONUPDATEMESSAGE = _descriptor.Descriptor(
    name="ObjectSearchPermissionUpdateMessage",
    full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_verify_key",
            full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage.target_verify_key",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_object_id",
            full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage.target_object_id",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="add_instead_of_remove",
            full_name="syft.core.node.common.service.ObjectSearchPermissionUpdateMessage.add_instead_of_remove",
            index=4,
            number=5,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=180,
    serialized_end=403,
)

_OBJECTSEARCHPERMISSIONUPDATEMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_OBJECTSEARCHPERMISSIONUPDATEMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_OBJECTSEARCHPERMISSIONUPDATEMESSAGE.fields_by_name[
    "target_object_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
DESCRIPTOR.message_types_by_name[
    "ObjectSearchPermissionUpdateMessage"
] = _OBJECTSEARCHPERMISSIONUPDATEMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ObjectSearchPermissionUpdateMessage = _reflection.GeneratedProtocolMessageType(
    "ObjectSearchPermissionUpdateMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _OBJECTSEARCHPERMISSIONUPDATEMESSAGE,
        "__module__": "proto.core.node.common.service.object_search_permission_update_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.service.ObjectSearchPermissionUpdateMessage)
    },
)
_sym_db.RegisterMessage(ObjectSearchPermissionUpdateMessage)


# @@protoc_insertion_point(module_scope)



# stdlib
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pandas as pd

# syft relative
from ....logger import traceback_and_raise
from ....util import validate_field
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ...io.route import Route
from ..common.client import Client
from .service import RequestMessage


class RequestQueueClient:
    def __init__(self, client: Client) -> None:
        self.client = client
        self.handlers = RequestHandlerQueueClient(client=client)

    @property
    def requests(self) -> List[RequestMessage]:
        # syft absolute
        from syft.core.node.domain.service.get_all_requests_service import (
            GetAllRequestsMessage,
        )

        msg = GetAllRequestsMessage(
            address=self.client.address, reply_to=self.client.address
        )

        requests = self.client.send_immediate_msg_with_reply(msg=msg).requests  # type: ignore

        for request in requests:
            request.gc_enabled = False
            request.owner_client_if_available = self.client

        return requests

    def get_request_id_from_object_id(self, object_id: UID) -> Optional[UID]:
        for req in self.requests:
            if req.object_id == object_id:
                return req.request_id

        return object_id

    def __getitem__(self, key: Union[str, int]) -> RequestMessage:
        if isinstance(key, str):
            for request in self.requests:
                if key == str(request.id.value):
                    return request
            traceback_and_raise(
                KeyError("No such request found for string id:" + str(key))
            )
        if isinstance(key, int):
            return self.requests[key]
        else:
            traceback_and_raise(KeyError("Please pass in a string or int key"))

        raise Exception("should not get here")

    def __repr__(self) -> str:
        return repr(self.requests)

    @property
    def pandas(self) -> pd.DataFrame:
        request_lines = [
            {
                "Requested Object's tags": request.object_tags,
                "Reason": request.request_description,
                "Request ID": request.id,
                "Requested Object's ID": request.object_id,
                "Requested Object's type": request.object_type,
            }
            for request in self.requests
        ]
        return pd.DataFrame(request_lines)

    def add_handler(
        self,
        action: str,
        print_local: bool = False,
        log_local: bool = False,
        tags: Optional[List[str]] = None,
        timeout_secs: int = -1,
        element_quota: Optional[int] = None,
    ) -> None:
        handler_opts = self._validate_options(
            id=UID(),
            action=action,
            print_local=print_local,
            log_local=log_local,
            tags=tags,
            timeout_secs=timeout_secs,
            element_quota=element_quota,
        )

        self._update_handler(handler_opts, keep=True)

    def remove_handler(self, key: Union[str, int]) -> None:
        handler_opts = self.handlers[key]

        self._update_handler(handler_opts, keep=False)

    def clear_handlers(self) -> None:
        for handler in self.handlers.handlers:
            id_str = str(handler["id"].value).replace("-", "")
            self.remove_handler(id_str)

    def _validate_options(
        self,
        action: str,
        print_local: bool = False,
        log_local: bool = False,
        tags: Optional[List[str]] = None,
        timeout_secs: int = -1,
        element_quota: Optional[int] = None,
        id: Optional[UID] = None,
    ) -> Dict[str, Any]:
        handler_opts: Dict[str, Any] = {}
        if action not in ["accept", "deny"]:
            traceback_and_raise(Exception("Action must be 'accept' or 'deny'"))
        handler_opts["action"] = action
        handler_opts["print_local"] = bool(print_local)
        handler_opts["log_local"] = bool(log_local)

        handler_opts["tags"] = []
        if tags is not None:
            for tag in tags:
                handler_opts["tags"].append(tag)
        handler_opts["timeout_secs"] = max(-1, int(timeout_secs))
        if element_quota is not None:
            handler_opts["element_quota"] = max(0, int(element_quota))

        if id is None:
            id = UID()
        handler_opts["id"] = id

        return handler_opts

    def _update_handler(self, request_handler: Dict[str, Any], keep: bool) -> None:
        # syft absolute
        from syft.core.node.domain.service.request_handler_service import (
            UpdateRequestHandlerMessage,
        )

        msg = UpdateRequestHandlerMessage(
            address=self.client.address, handler=request_handler, keep=keep
        )
        self.client.send_immediate_msg_without_reply(msg=msg)


class RequestHandlerQueueClient:
    def __init__(self, client: Client) -> None:
        self.client = client

    @property
    def handlers(self) -> List[Dict]:
        # syft absolute
        from syft.core.node.domain.service.request_handler_service import (
            GetAllRequestHandlersMessage,
        )

        msg = GetAllRequestHandlersMessage(
            address=self.client.address, reply_to=self.client.address
        )
        return validate_field(
            self.client.send_immediate_msg_with_reply(msg=msg), "handlers"
        )

    def __getitem__(self, key: Union[str, int]) -> Dict:
        """
        allow three ways to get an request handler:
            1. use id: str
            2. use tag: str
            3. use index: int
        """
        if isinstance(key, str):
            matches = 0
            match_handler: Optional[Dict] = None
            for handler in self.handlers:
                if key in str(handler["id"].value).replace("-", ""):
                    return handler
                if key in handler["tags"]:
                    matches += 1
                    match_handler = handler
            if matches == 1 and match_handler is not None:
                return match_handler
            elif matches > 1:
                raise KeyError("More than one item with tag:" + str(key))

            raise KeyError("No such request found for string id:" + str(key))
        if isinstance(key, int):
            return self.handlers[key]
        else:
            raise KeyError("Please pass in a string or int key")

    def __repr__(self) -> str:
        return repr(self.handlers)

    @property
    def pandas(self) -> pd.DataFrame:
        def _get_time_remaining(handler: dict) -> int:
            timeout_secs = handler.get("timeout_secs", -1)
            if timeout_secs == -1:
                return -1
            else:
                created_time = handler.get("created_time", 0)
                rem = timeout_secs - (time.time() - created_time)
                return round(rem)

        handler_lines = [
            {
                "tags": handler["tags"],
                "ID": handler["id"],
                "action": handler["action"],
                "remaining time (s):": _get_time_remaining(handler),
            }
            for handler in self.handlers
        ]
        return pd.DataFrame(handler_lines)


class DomainClient(Client):

    domain: SpecificLocation
    requests: RequestQueueClient

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        domain: SpecificLocation,
        network: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
    ):
        super().__init__(
            name=name,
            routes=routes,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
        )

        self.requests = RequestQueueClient(client=self)
        self.post_init()

    @property
    def id(self) -> UID:
        return self.domain.id

    @property
    def device(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself, this property will return the Location of that device
        if it is known by the client."""

        return super().device

    @device.setter
    def device(self, new_device: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself and we learn the Location of that device, this setter
        allows us to save the Location of that device for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a domain, you don't need a Device ID.")
        )

    @property
    def vm(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself, this property will return the Location of that vm
        if it is known by the client."""

        return super().vm

    @vm.setter
    def vm(self, new_vm: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself and we learn the Location of that vm, this setter
        allows us to save the Location of that vm for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a device, you don't need a VM Location.")
        )

    def __repr__(self) -> str:
        no_dash = str(self.id).replace("-", "")
        return f"<{type(self).__name__}: {no_dash}>"

    def update_vars(self, state: dict) -> pd.DataFrame:
        for ptr in self.store.store:
            tags = getattr(ptr, "tags", None)
            if tags is not None:
                for tag in tags:
                    state[tag] = ptr
        return self.store.pandas



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/device.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/torch/device.proto",
    package="syft.lib.torch",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1cproto/lib/torch/device.proto\x12\x0esyft.lib.torch"%\n\x06\x44\x65vice\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x05\x62\x06proto3',
)


_DEVICE = _descriptor.Descriptor(
    name="Device",
    full_name="syft.lib.torch.Device",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="type",
            full_name="syft.lib.torch.Device.type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="index",
            full_name="syft.lib.torch.Device.index",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=48,
    serialized_end=85,
)

DESCRIPTOR.message_types_by_name["Device"] = _DEVICE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Device = _reflection.GeneratedProtocolMessageType(
    "Device",
    (_message.Message,),
    {
        "DESCRIPTOR": _DEVICE,
        "__module__": "proto.lib.torch.device_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.Device)
    },
)
_sym_db.RegisterMessage(Device)


# @@protoc_insertion_point(module_scope)



# stdlib
import glob
import os
import time

# third party
from model import SNLIClassifier
import torch
import torch.nn as nn
import torch.optim as O
from torchtext import data
from torchtext import datasets
from util import get_args
from util import makedirs

args = get_args()
if torch.cuda.is_available():
    torch.cuda.set_device(args.gpu)
    device = torch.device(f"cuda:{args.gpu}")
else:
    device = torch.device("cpu")

inputs = data.Field(lower=args.lower, tokenize="spacy")
answers = data.Field(sequential=False)

train, dev, test = datasets.SNLI.splits(inputs, answers)

inputs.build_vocab(train, dev, test)
if args.word_vectors:
    if os.path.isfile(args.vector_cache):
        inputs.vocab.vectors = torch.load(args.vector_cache)
    else:
        inputs.vocab.load_vectors(args.word_vectors)
        makedirs(os.path.dirname(args.vector_cache))
        torch.save(inputs.vocab.vectors, args.vector_cache)
answers.build_vocab(train)

train_iter, dev_iter, test_iter = data.BucketIterator.splits(
    (train, dev, test), batch_size=args.batch_size, device=device
)

config = args
config.n_embed = len(inputs.vocab)
config.d_out = len(answers.vocab)
config.n_cells = config.n_layers

# double the number of cells for bidirectional networks
if config.birnn:
    config.n_cells *= 2

if args.resume_snapshot:
    model = torch.load(args.resume_snapshot, map_location=device)
else:
    model = SNLIClassifier(config)
    if args.word_vectors:
        model.embed.weight.data.copy_(inputs.vocab.vectors)
        model.to(device)

criterion = nn.CrossEntropyLoss()
opt = O.Adam(model.parameters(), lr=args.lr)

iterations = 0
start = time.time()
best_dev_acc = -1
header = "  Time Epoch Iteration Progress    (%Epoch)   Loss   Dev/Loss     Accuracy  Dev/Accuracy"
dev_log_template = " ".join(
    "{:>6.0f},{:>5.0f},{:>9.0f},{:>5.0f}/{:<5.0f} {:>7.0f}%,{:>8.6f},{:8.6f},{:12.4f},{:12.4f}".split(
        ","
    )
)
log_template = " ".join(
    "{:>6.0f},{:>5.0f},{:>9.0f},{:>5.0f}/{:<5.0f} {:>7.0f}%,{:>8.6f},{},{:12.4f},{}".split(
        ","
    )
)
makedirs(args.save_path)
print(header)

for epoch in range(args.epochs):
    train_iter.init_epoch()
    n_correct, n_total = 0, 0
    for batch_idx, batch in enumerate(train_iter):

        # switch model to training mode, clear gradient accumulators
        model.train()
        opt.zero_grad()

        iterations += 1

        # forward pass
        answer = model(batch)

        # calculate accuracy of predictions in the current batch
        n_correct += (
            (torch.max(answer, 1)[1].view(batch.label.size()) == batch.label)
            .sum()
            .item()
        )
        n_total += batch.batch_size
        train_acc = 100.0 * n_correct / n_total

        # calculate the loss of the network output with respect to training labels
        loss = criterion(answer, batch.label)

        # backpropagate and update optimizer learning rate
        loss.backward()
        opt.step()

        # checkpoint model periodically
        if iterations % args.save_every == 0:
            snapshot_prefix = os.path.join(args.save_path, "snapshot")
            snapshot_path = (
                snapshot_prefix
                + "_acc_{:.4f}_loss_{:.6f}_iter_{}_model.pt".format(
                    train_acc, loss.item(), iterations
                )
            )
            torch.save(model, snapshot_path)
            for f in glob.glob(snapshot_prefix + "*"):
                if f != snapshot_path:
                    os.remove(f)

        # evaluate performance on validation set periodically
        if iterations % args.dev_every == 0:

            # switch model to evaluation mode
            model.eval()
            dev_iter.init_epoch()

            # calculate accuracy on validation set
            n_dev_correct, dev_loss = 0, 0
            with torch.no_grad():
                for dev_batch_idx, dev_batch in enumerate(dev_iter):
                    answer = model(dev_batch)
                    n_dev_correct += (
                        (
                            torch.max(answer, 1)[1].view(dev_batch.label.size())
                            == dev_batch.label
                        )
                        .sum()
                        .item()
                    )
                    dev_loss = criterion(answer, dev_batch.label)
            dev_acc = 100.0 * n_dev_correct / len(dev)

            print(
                dev_log_template.format(
                    time.time() - start,
                    epoch,
                    iterations,
                    1 + batch_idx,
                    len(train_iter),
                    100.0 * (1 + batch_idx) / len(train_iter),
                    loss.item(),
                    dev_loss.item(),
                    train_acc,
                    dev_acc,
                )
            )

            # update best valiation set accuracy
            if dev_acc > best_dev_acc:

                # found a model with better validation set accuracy

                best_dev_acc = dev_acc
                snapshot_prefix = os.path.join(args.save_path, "best_snapshot")
                snapshot_path = (
                    snapshot_prefix
                    + "_devacc_{}_devloss_{}__iter_{}_model.pt".format(
                        dev_acc, dev_loss.item(), iterations
                    )
                )

                # save model, delete previous 'best_snapshot' files
                torch.save(model, snapshot_path)
                for f in glob.glob(snapshot_prefix + "*"):
                    if f != snapshot_path:
                        os.remove(f)

        elif iterations % args.log_every == 0:

            # print progress message
            print(
                log_template.format(
                    time.time() - start,
                    epoch,
                    iterations,
                    1 + batch_idx,
                    len(train_iter),
                    100.0 * (1 + batch_idx) / len(train_iter),
                    loss.item(),
                    " " * 8,
                    n_correct / n_total * 100,
                    " " * 12,
                )
            )
        if args.dry_run:
            break



# stdlib
import copy
import os
from pathlib import Path
import time
from typing import Any
from typing import Tuple

# third party
import pytest
import torch

# syft absolute
import syft as sy

IN_DIM = 100
OUT_DIM = 10


class SyNet(sy.Module):
    """
    Simple test model
    """

    def __init__(self) -> None:
        super(SyNet, self).__init__(torch_ref=torch)
        self.fc1 = torch.nn.Linear(IN_DIM, OUT_DIM)

    def forward(self, x: torch.Tensor) -> Any:
        return self.fc1(x)


class SyNetEmpty(sy.Module):
    """
    Simple test model
    """

    def __init__(self) -> None:
        super(SyNetEmpty, self).__init__(torch_ref=torch)

    def forward(self, x: torch.Tensor) -> Any:
        return 0


@pytest.fixture(scope="function")
def model() -> SyNet:
    return SyNet()


@pytest.fixture(scope="function")
def modelEmpty() -> SyNetEmpty:
    return SyNetEmpty()


@pytest.fixture(scope="function")
def dataloader() -> Tuple[torch.Tensor, torch.Tensor]:
    return torch.randn(size=(1, IN_DIM)), torch.randn(size=(1, OUT_DIM))


def test_repr_to_kwargs() -> None:
    assert sy.lib.util.full_name_with_qualname(klass=torch.Tensor) == "torch.Tensor"
    assert sy.lib.torch.module.repr_to_kwargs(
        "1, 32, kernel_size=(3, 3), stride=(1, 1)"
    ) == ([1, 32], {"kernel_size": (3, 3), "stride": (1, 1)})
    assert sy.lib.torch.module.repr_to_kwargs("1, 32") == ([1, 32], {})
    assert sy.lib.torch.module.repr_to_kwargs("kernel_size=(3, 3), stride=(1, 1)") == (
        [],
        {"kernel_size": (3, 3), "stride": (1, 1)},
    )


def test_module_setup(root_client: sy.VirtualMachineClient, model: SyNet) -> None:
    remote = copy.copy(model)
    remote.setup(torch_ref=root_client.torch)
    assert remote.is_local is False
    assert remote.torch_ref == root_client.torch
    assert remote.training is False

    remote.setup(torch_ref=torch)
    assert remote.is_local is True


def test_module_attr(model: SyNet) -> None:
    model.__setattr__("fc1", torch.nn.Linear(1, 2))
    assert model.__getattr__("fc1").in_features == 1
    assert model.__getattr__("fc1").out_features == 2


def test_module_modules(model: SyNet) -> None:
    modules = model.modules
    assert len(modules.items()) == 1
    assert "fc1" in modules
    assert modules["fc1"].in_features == IN_DIM


def test_module_modules_empty(modelEmpty: SyNetEmpty) -> None:
    modules = modelEmpty.modules
    assert len(modules.items()) == 0


@pytest.mark.slow
def test_module_parameteres(root_client: sy.VirtualMachineClient, model: SyNet) -> None:
    model_ptr = model.send(root_client)

    assert len(model_ptr.parameters().get()) == 2
    assert model_ptr.parameters().get()[0].shape == torch.Size([OUT_DIM, IN_DIM])
    assert model_ptr.parameters().get()[1].shape == torch.Size([OUT_DIM])


def test_module_cuda(model: SyNet) -> None:
    model.cpu()
    assert model.parameters()[-1].is_cuda is False


def test_module_zero(model: SyNet) -> None:
    model.zero_layers()
    for _, m in model.modules.items():
        for _, v in m.state_dict().items():
            assert v.sum() == 0


def test_module_state_dict(model: SyNet) -> None:
    state = model.state_dict()

    new_model = SyNet()
    new_model.load_state_dict(state)

    new_state = new_model.state_dict()
    for k in state:
        assert k in new_state
        assert torch.all(torch.eq(new_state[k], state[k]))

    new_model.is_local = False

    assert new_model.load_state_dict(state) is None
    assert new_model.state_dict() is None


def test_module_load_save(model: SyNet) -> None:
    state = model.state_dict()

    folder = Path("tmp")
    try:
        os.mkdir(folder)
    except FileExistsError:
        pass

    path = folder / str(time.time())
    model.save(path)

    model.is_local = False
    assert model.save(path) is None

    new_model = SyNet()
    new_model.load(path)
    new_state = new_model.state_dict()

    new_model.is_local = False
    assert new_model.load(path) is None

    try:
        os.remove(path)
    except BaseException:
        pass

    for k in state:
        assert k in new_state
        assert torch.all(torch.eq(new_state[k], state[k]))


def test_module_gradient_sanity(
    model: SyNet,
    dataloader: Tuple[torch.Tensor, torch.Tensor],
) -> None:
    data, labels = dataloader

    result = model(data)
    loss_func = torch.nn.L1Loss()
    loss = loss_func(result, labels)
    loss.backward()

    assert model.parameters()[-1].grad is not None


@pytest.mark.slow
def test_module_send_get(
    root_client: sy.VirtualMachineClient,
    model: SyNet,
    dataloader: Tuple[torch.Tensor, torch.Tensor],
) -> None:
    data, labels = dataloader

    model_ptr = model.send(root_client)
    data_ptr = data.send(root_client)
    labels_ptr = labels.send(root_client)

    results_ptr = model_ptr(data_ptr)
    remote_loss_func = root_client.torch.nn.L1Loss()
    remote_loss = remote_loss_func(results_ptr, labels_ptr)
    remote_loss.backward()

    direct_param = model_ptr.parameters().get()
    for param in direct_param:
        assert param.grad is not None

    # get() uses state_dict/load_state_dict
    # load_state_dict breaks the computational graph, and we won't have the gradients here.
    # ref: https://discuss.pytorch.org/t/loading-a-state-dict-seems-to-erase-grad/56676
    model_parameter = model_ptr.get().parameters()
    for param in model_parameter:
        assert param.grad is None

    for idx, param in enumerate(direct_param):
        assert param.tolist() == model_parameter[idx].tolist()

    assert model.get() is None

    model.is_local = False
    assert model.send(root_client) is None


@pytest.mark.slow
def test_debug_sum_layers(root_client: sy.VirtualMachineClient, model: SyNet) -> None:
    assert model.debug_sum_layers() is None
    model_ptr = model.send(root_client)

    assert model_ptr.debug_sum_layers() is None



# stdlib
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy

# syft relative
from .utils_test import decrypt

ts = pytest.importorskip("tenseal")
sy.load("tenseal")


def _almost_equal(vec1: Any, vec2: Any, precision_pow_ten: int = 1) -> None:
    if isinstance(vec1, ts.PlainTensor):
        vec1 = vec1.tolist()
    if isinstance(vec2, ts.PlainTensor):
        vec2 = vec2.tolist()

    upper_bound = pow(10, -precision_pow_ten)
    assert pytest.approx(vec1, abs=upper_bound) == vec2


@pytest.fixture(scope="function")
def context() -> Any:
    context = ts.context(
        ts.SCHEME_TYPE.CKKS, 8192, coeff_mod_bit_sizes=[60, 40, 40, 60]
    )
    context.global_scale = pow(2, 40)
    return context


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_sanity(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    enc_v1 = ts.ckks_tensor(context, v1)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, [0, 1, 2, 3, 4])


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_add(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 + v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)
    enc_v2 = ts.ckks_tensor(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # add
    result_enc_ptr = enc_v1_ptr + enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)

    # add inplace
    enc_v1_ptr += enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_sub(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 - v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)
    enc_v2 = ts.ckks_tensor(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # sub
    result_enc_ptr = enc_v1_ptr - enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)
    _almost_equal(result, expected)

    # sub inplace
    enc_v1_ptr -= enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_mul(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 * v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)
    enc_v2 = ts.ckks_tensor(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # mul
    result_enc_ptr = enc_v1_ptr * enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)

    # mul inplace
    enc_v1_ptr *= enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_iadd(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 + v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)

    v2_ptr = sy.lib.python.List(v2).send(root_client, pointable=True)
    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # iadd
    result_enc_ptr = enc_v1_ptr + v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)

    # iadd inplace
    enc_v1_ptr += v2_ptr

    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_isub(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 - v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)

    v2_ptr = sy.lib.python.List(v2).send(root_client, pointable=True)
    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # isub
    result_enc_ptr = enc_v1_ptr - v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_imul(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 * v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.ckks_tensor(context, v1)

    v2_ptr = sy.lib.python.List(v2).send(root_client, pointable=True)
    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # imul
    result_enc_ptr = enc_v1_ptr * v2_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, expected)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_power(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    enc_v1 = ts.ckks_tensor(context, [0, 1, 2, 3, 4])

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result_enc_ptr = enc_v1_ptr ** 3

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, [0, 1, 8, 27, 64])


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_negation(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    enc_v1 = ts.ckks_tensor(context, [1, 2, 3, 4, 5])

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result_enc_ptr = -enc_v1_ptr

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, [-1, -2, -3, -4, -5])


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_square(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    enc_v1 = ts.ckks_tensor(context, [0, 1, 2, 3, 4])

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result_enc_ptr = enc_v1_ptr.square()

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, [0, 1, 4, 9, 16])


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_sum(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    enc_v1 = ts.ckks_tensor(context, [0, 1, 2, 3, 4])

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result_enc_ptr = enc_v1_ptr.sum()

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, 10)


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_polyval(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    polynom = [1, 2, 3, 4]
    enc_v1 = ts.ckks_tensor(context, [-2, 2])

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result_enc_ptr = enc_v1_ptr.polyval(polynom)

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, [-23, 49])


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_ckkstensor_dot(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]

    enc_v1 = ts.ckks_tensor(context, v1)
    enc_v2 = ts.ckks_tensor(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    result_enc_ptr = enc_v1_ptr.dot(enc_v2_ptr)

    result = decrypt(context, result_enc_ptr)
    _almost_equal(result, 10)

    # inplace

    enc_v1_ptr.dot_(enc_v2_ptr)
    result = decrypt(context, enc_v1_ptr)
    _almost_equal(result, 10)



# stdlib
import asyncio
import json
from typing import Any
from unittest.mock import Mock
from unittest.mock import patch

# third party
from aiortc import RTCDataChannel
from aiortc import RTCPeerConnection
from aiortc import RTCSessionDescription
from aiortc.contrib.signaling import object_from_string
from nacl.signing import SigningKey
import pytest
from pytest import MonkeyPatch

# syft absolute
from syft import serialize
from syft.core.node.common.service.repr_service import ReprMessage
from syft.core.node.domain.domain import Domain
from syft.grid.connections.webrtc import DC_CHUNK_START_SIGN
from syft.grid.connections.webrtc import DC_MAX_CHUNK_SIZE
from syft.grid.connections.webrtc import OrderedChunk
from syft.grid.connections.webrtc import WebRTCConnection


class AsyncMock(Mock):
    async def __call__(self, *args: Any, **kwargs: Any) -> None:
        return super(AsyncMock, self).__call__(*args, **kwargs)


def get_signing_key() -> SigningKey:
    # return a the signing key used to sign the get_signed_message_bytes fixture
    key = "e89ff2e651b42393b6ecb5956419088781309d953d72bd73a0968525a3a6a951"
    return SigningKey(bytes.fromhex(key))


# --------------------- INIT ---------------------


@pytest.mark.asyncio
async def test_init() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    assert webrtc is not None
    assert webrtc.node == domain
    assert webrtc.loop is not None
    assert isinstance(webrtc.producer_pool, asyncio.Queue)
    assert isinstance(webrtc.consumer_pool, asyncio.Queue)
    assert isinstance(webrtc.peer_connection, RTCPeerConnection)
    assert not webrtc._client_address


@pytest.mark.asyncio
async def test_init_raise_exception(monkeypatch: MonkeyPatch) -> None:
    with patch("syft.grid.connections.webrtc.traceback_and_raise") as mock_logger:
        with patch(
            "syft.grid.connections.webrtc.RTCPeerConnection", side_effect=Exception()
        ):
            domain = Domain(name="test")
            WebRTCConnection(node=domain)
            assert mock_logger.assert_called


# --------------------- METHODS ---------------------


@pytest.mark.asyncio
async def test_set_offer_raise_exception() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)

    with patch(
        "syft.grid.connections.webrtc.RTCPeerConnection.createDataChannel",
        side_effect=Exception(),
    ):
        with pytest.raises(Exception):
            await webrtc._set_offer()


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_offer_sets_channel() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    await webrtc._set_offer()
    assert isinstance(webrtc.channel, RTCDataChannel)
    assert webrtc.channel.bufferedAmountLowThreshold == 4 * DC_MAX_CHUNK_SIZE


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_offer_on_open() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    await webrtc._set_offer()

    channel_methods = list(webrtc.channel._events.values())
    on_open = list(channel_methods[1].values())[0]

    coro_mock = AsyncMock()
    with patch(
        "syft.grid.connections.webrtc.WebRTCConnection.producer",
        return_value=coro_mock(),
    ) as producer_mock:
        await on_open()
        assert producer_mock.call_count == 1


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_offer_on_message() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    await webrtc._set_offer()

    channel_methods = list(webrtc.channel._events.values())
    on_message = list(channel_methods[2].values())[0]

    coro_mock = AsyncMock()
    with patch(
        "syft.grid.connections.webrtc.WebRTCConnection.consumer",
        return_value=coro_mock(),
    ) as consumer_mock:
        await on_message(OrderedChunk(1, DC_CHUNK_START_SIGN).save())
        assert consumer_mock.call_count == 0

        await on_message(OrderedChunk(0, b"a").save())
        assert consumer_mock.call_count == 1


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_answer_raise_exception() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    offer_payload = await webrtc._set_offer()

    # FIXME: Nahua is not happy with this test because it "indirectly" triggered exception
    # https://github.com/OpenMined/PySyft/issues/5126
    with patch("syft.grid.connections.webrtc.traceback_and_raise") as mock_logger:
        with pytest.raises(Exception):
            # This would fail because 'have-local-offer' is applied
            await webrtc._set_answer(payload=offer_payload)
        assert mock_logger.called


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_answer_on_datachannel() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    offer_payload = await webrtc._set_offer()

    answer_webrtc = WebRTCConnection(node=domain)
    await answer_webrtc._set_answer(payload=offer_payload)

    channel_methods = list(answer_webrtc.peer_connection._events.values())
    on_datachannel = list(channel_methods[1].values())[0]

    coro_mock = AsyncMock()
    with patch(
        "syft.grid.connections.webrtc.WebRTCConnection.producer",
        return_value=coro_mock(),
    ) as producer_mock:
        channel = answer_webrtc.peer_connection.createDataChannel("datachannel")
        on_datachannel(channel)
        assert producer_mock.call_count == 1


@pytest.mark.slow
@pytest.mark.asyncio
async def test_set_answer_on_message() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    offer_payload = await webrtc._set_offer()

    answer_webrtc = WebRTCConnection(node=domain)
    await answer_webrtc._set_answer(payload=offer_payload)

    channel_methods = list(answer_webrtc.peer_connection._events.values())
    on_channel = list(channel_methods[1].values())[0]

    coro_mock = AsyncMock()
    with patch(
        "syft.grid.connections.webrtc.WebRTCConnection.consumer",
        return_value=coro_mock(),
    ) as consumer_mock:
        channel = answer_webrtc.peer_connection.createDataChannel("datachannel")
        on_channel(channel)

        channel_methods = list(answer_webrtc.channel._events.values())
        on_message = list(channel_methods[1].values())[0]

        await on_message(OrderedChunk(1, DC_CHUNK_START_SIGN).save())
        assert consumer_mock.call_count == 0

        await on_message(OrderedChunk(0, b"a").save())
        assert consumer_mock.call_count == 1


@pytest.mark.asyncio
async def test_finish_coroutines_raise_exception() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)

    with patch("syft.grid.connections.webrtc.traceback_and_raise") as mock_logger:
        with patch(
            "syft.grid.connections.webrtc.RTCDataChannel.close", side_effect=Exception()
        ):
            webrtc._finish_coroutines()
            assert mock_logger.called


@pytest.mark.asyncio
async def test_close_raise_exception() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)

    with patch("syft.grid.connections.webrtc.traceback_and_raise") as mock_logger:
        with patch(
            "syft.grid.connections.webrtc.RTCDataChannel.close", side_effect=Exception()
        ):
            webrtc.close()
            assert mock_logger.called


@pytest.mark.slow
@pytest.mark.asyncio
async def test_close() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    await webrtc._set_offer()

    with patch("syft.grid.connections.webrtc.RTCDataChannel.send") as send_mock:
        with patch(
            "syft.grid.connections.webrtc.WebRTCConnection._finish_coroutines"
        ) as finish_mock:
            webrtc.close()
            assert send_mock.call_count == 1
            assert finish_mock.call_count == 1


# --------------------- INTEGRATION ---------------------


@pytest.mark.asyncio
def test_init_without_event_loop() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)
    assert webrtc is not None


@pytest.mark.slow
@pytest.mark.asyncio
async def test_signaling_process() -> None:
    domain = Domain(name="test")
    webrtc = WebRTCConnection(node=domain)

    offer_payload = await webrtc._set_offer()
    offer_dict = json.loads(offer_payload)
    aiortc_session = object_from_string(offer_payload)

    assert "sdp" in offer_dict
    assert "type" in offer_dict
    assert offer_dict["type"] == "offer"
    assert isinstance(aiortc_session, RTCSessionDescription)

    answer_webrtc = WebRTCConnection(node=domain)
    answer_payload = await answer_webrtc._set_answer(payload=offer_payload)
    answer_dict = json.loads(answer_payload)
    aiortc_session = object_from_string(answer_payload)

    assert "sdp" in answer_dict
    assert "type" in answer_dict
    assert answer_dict["type"] == "answer"
    assert isinstance(aiortc_session, RTCSessionDescription)

    response = await webrtc._process_answer(payload=answer_payload)
    assert response is None


@pytest.mark.asyncio
async def test_consumer_request() -> None:
    test_domain = Domain(name="test")

    webrtc_node = WebRTCConnection(node=test_domain)

    msg = ReprMessage(address=test_domain.address)
    signing_key = SigningKey.generate()
    test_domain.root_verify_key = signing_key.verify_key
    signed_msg = msg.sign(signing_key=signing_key)

    msg_bin = serialize(signed_msg, to_bytes=True)

    await webrtc_node.consumer(msg=msg_bin)



# stdlib
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy


def get_permission(
    obj: Any, node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    remote_obj = node.store[obj.id_at_location]
    remote_obj.read_permissions[client.verify_key] = obj.id_at_location


inputs = {
    "__add__": [[[1, 2, 3]], [[1, 2.5, True]]],
    "__contains__": [[41], [-1], [None], ["test"]],
    "__eq__": [[[41, 15, 3, 80]], [[True]]],
    "__ge__": [[[41, 15, 3, 80]], [[True]]],
    "__getitem__": [[1], [2], [3]],
    "__gt__": [[[41, 15, 3, 80]], [[True]]],
    "__iadd__": [[[1, 2, 3]], [[1, 2.5, True]]],
    "__imul__": [[1], [3], [5]],
    "__le__": [[[41, 15, 3, 80]], [[True]]],
    "__len__": [[]],
    "__lt__": [[[41, 15, 3, 80]], [[True]]],
    "__mul__": [[1], [3], [5]],
    "__ne__": [[[41, 15, 3, 80]], [[True]]],
    "__rmul__": [[1], [3], [5]],
    "__setitem__": [[0, 2], [1, 5]],
    "append": [[1], [2], [3]],
    "clear": [[]],
    "copy": [[]],
    "count": [[1], [2]],
    "extend": [[[1, 2, 3]], [[4, 5, 6]]],
    "index": [[0], [1], [5]],
    "insert": [[0, "a"], [3, "b"]],
    "pop": [[0], [3]],
    "remove": [[1], [42]],
    "reverse": [[]],
    "sort": [[]],
}


def generate_0():
    return ([41, 15, 3, 80],)


def generate_1():
    return (list(range(2 ** 5)),)


mapping = {0: generate_0, 1: generate_1}

objects = [0, 1]


@pytest.mark.xfail
@pytest.mark.slow
@pytest.mark.parametrize("test_object", objects)
@pytest.mark.parametrize("func", inputs.keys())
def test_pointer_objectives(
    test_object, func, node: sy.VirtualMachine, client: sy.VirtualMachineClient
):
    py_obj = mapping[test_object]()
    sy_obj, remote_sy_obj = sy.lib.python.List(py_obj), client.syft.lib.python.List(
        py_obj
    )
    possible_inputs = inputs[func]

    if not hasattr(py_obj, func):
        return

    py_method = getattr(py_obj, func)
    sy_method = getattr(sy_obj, func)

    if func == "__len__":
        func = "len"

    remote_sy_method = getattr(remote_sy_obj, func)

    for possible_input in possible_inputs:
        try:
            py_res = py_method(*possible_input)
        except Exception as py_e:
            py_res = str(py_e)

        try:
            sy_res = sy_method(*possible_input)
        except Exception as sy_e:
            sy_res = str(sy_e)

        try:
            remote_sy_res = remote_sy_method(*possible_input)
            get_permission(remote_sy_res, node, client)
            remote_sy_res = remote_sy_res.get()
        except Exception as remote_sy_e:
            remote_sy_res = str(remote_sy_e)
        #
        # if isinstance(py_res, float):
        #     py_res = int(py_res * 1000) / 1000
        #     sy_res = int(sy_res * 1000) / 1000
        #     remote_sy_res = int(remote_sy_res * 1000) / 1000

        assert py_res == sy_res
        assert sy_res == remote_sy_res

        assert py_obj == sy_obj
        # TODO add this as well when the store logic will work
        get_permission(remote_sy_obj, node, client)
        assert sy_obj == remote_sy_obj.get()


@pytest.mark.slow
@pytest.mark.parametrize("test_object", objects)
def test_iterator(test_object, client: sy.VirtualMachineClient):
    py_obj = mapping[test_object]()
    sy_obj, remote_sy_obj = sy.lib.python.List(py_obj), client.syft.lib.python.List(
        py_obj
    )

    py_iter = iter(py_obj)
    sy_iter = iter(sy_obj)

    remote_sy_obj.set_request_config({})
    rsy_iter = iter(remote_sy_obj)

    for i in range(len(py_obj)):
        py_elem = next(py_iter)
        sy_elem = next(sy_iter)
        rsy_elem = next(rsy_iter)

        assert py_elem == sy_elem
        assert sy_elem == rsy_elem.get()


@pytest.mark.slow
@pytest.mark.parametrize("test_object", objects)
def test_reversed_iterator(test_object, client):
    py_obj = mapping[test_object]()
    sy_obj, remote_sy_obj = sy.lib.python.List(py_obj), client.syft.lib.python.List(
        py_obj
    )

    py_iter = reversed(py_obj)
    sy_iter = reversed(sy_obj)
    rsy_iter = reversed(remote_sy_obj)

    for i in range(len(py_obj)):
        py_elem = next(py_iter)
        sy_elem = next(sy_iter)
        rsy_elem = next(rsy_iter)

        assert py_elem == sy_elem
        assert sy_elem == rsy_elem.get()



# stdlib
from dataclasses import dataclass
from typing import Optional

# syft relative
from .specs.cpu import CPU
from .specs.gpu import GPU
from .specs.network import Network
from .specs.provider import Provider
from .specs.storage import Storage


@dataclass
class DeviceType:
    name: str
    provider: Provider
    price_per_hour: float
    memory: int
    bare_metal: bool
    storage: Storage
    cpu: CPU
    gpu: Optional[GPU]
    network: Optional[Network]
    spot_mode_supported: bool = False
    price_per_hour_spot: float = 0



# stdlib
import dataclasses
from uuid import UUID

# third party
from sympc.config import Config
from sympc.session import Session
from sympc.store import CryptoStore

# syft relative
from ...proto.lib.sympc.session_pb2 import MPCSession as MPCSession_PB
from ..python import Dict
from ..python.primitive_factory import PrimitiveFactory


def protobuf_session_serializer(session: Session) -> MPCSession_PB:
    conf = PrimitiveFactory.generate_primitive(value=dataclasses.asdict(session.config))
    conf_proto = conf._object2proto()

    length_rs = session.ring_size.bit_length()
    rs_bytes = session.ring_size.to_bytes((length_rs + 7) // 8, byteorder="big")

    length_nr_parties = session.ring_size.bit_length()
    nr_parties_bytes = session.nr_parties.to_bytes(
        (length_nr_parties + 7) // 8, byteorder="big"
    )

    protocol = session.protocol.__name__
    protocol_serialized = str.encode(protocol)

    return MPCSession_PB(
        uuid=session.uuid.bytes,
        config=conf_proto,
        ring_size=rs_bytes,
        nr_parties=nr_parties_bytes,
        rank=session.rank,
        protocol=protocol_serialized,
    )


def protobuf_session_deserializer(proto: MPCSession_PB) -> Session:
    id_session = UUID(bytes=proto.uuid)
    rank = proto.rank
    conf_dict = Dict._proto2object(proto=proto.config)
    _conf_dict = {key: value for key, value in conf_dict.items()}
    conf = Config(**_conf_dict)
    ring_size = int.from_bytes(proto.ring_size, "big")
    nr_parties = int.from_bytes(proto.nr_parties, "big")
    protocol_deserialized = proto.protocol.decode()

    session = Session(
        config=conf,
        uuid=id_session,
        ring_size=ring_size,
        protocol=protocol_deserialized,
    )
    session.rank = rank
    session.crypto_store = CryptoStore()
    session.nr_parties = nr_parties

    return session



# future
from __future__ import print_function

# stdlib
import argparse
import os
import random

# third party
import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn
import torch.nn.parallel
import torch.optim as optim
import torch.utils.data
import torchvision.datasets as dset
import torchvision.transforms as transforms
import torchvision.utils as vutils

parser = argparse.ArgumentParser()
parser.add_argument(
    "--dataset",
    required=True,
    help="cifar10 | lsun | mnist |imagenet | folder | lfw | fake",
)
parser.add_argument("--data_folder", required=False, help="path to dataset")
parser.add_argument(
    "--data_workers", type=int, help="number of data loading data_workers", default=2
)
parser.add_argument("--batch_size", type=int, default=64, help="input batch size")
parser.add_argument(
    "--image_size",
    type=int,
    default=64,
    help="the height / width of the input image to network",
)
parser.add_argument("--nz", type=int, default=100, help="size of the latent z vector")
parser.add_argument("--ngf", type=int, default=64)
parser.add_argument("--ndf", type=int, default=64)
parser.add_argument(
    "--n_iter", type=int, default=25, help="number of epochs to train for"
)
parser.add_argument(
    "--lr", type=float, default=0.0002, help="learning rate, default=0.0002"
)
parser.add_argument(
    "--beta1", type=float, default=0.5, help="beta1 for adam. default=0.5"
)
parser.add_argument("--cuda", action="store_true", help="enables cuda")
parser.add_argument(
    "--dry-run", action="store_true", help="check a single training cycle works"
)
parser.add_argument("--n_gpu", type=int, default=1, help="number of GPUs to use")
parser.add_argument("--net_generator", default="", help="path to net_generator (to continue training)")
parser.add_argument("--net_discriminator", default="", help="path to net_discriminator (to continue training)")
parser.add_argument(
    "--out_folder", default=".", help="folder to output images and model checkpoints"
)
parser.add_argument("--manual_seed", type=int, help="manual seed")
parser.add_argument(
    "--classes",
    default="bedroom",
    help="comma separated list of classes for the lsun data set",
)

opt = parser.parse_args()
print(opt)

try:
    os.makedirs(opt.out_folder)
except OSError:
    pass

if opt.manual_seed is None:
    opt.manual_seed = random.randint(1, 10000)
print("Random Seed: ", opt.manual_seed)
random.seed(opt.manual_seed)
torch.manual_seed(opt.manual_seed)

cudnn.benchmark = True

if torch.cuda.is_available() and not opt.cuda:
    print("WARNING: You have a CUDA device, so you should probably run with --cuda")

if opt.data_folder is None and str(opt.dataset).lower() != "fake":
    raise ValueError(f'`data_folder` parameter is required for dataset "{opt.dataset}"')

if opt.dataset in ["imagenet", "folder", "lfw"]:
    # folder dataset
    dataset = dset.ImageFolder(
        root=opt.data_folder,
        transform=transforms.Compose(
            [
                transforms.Resize(opt.image_size),
                transforms.CenterCrop(opt.image_size),
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        ),
    )
    nc = 3
elif opt.dataset == "lsun":
    classes = [c + "_train" for c in opt.classes.split(",")]
    dataset = dset.LSUN(
        root=opt.data_folder,
        classes=classes,
        transform=transforms.Compose(
            [
                transforms.Resize(opt.image_size),
                transforms.CenterCrop(opt.image_size),
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        ),
    )
    nc = 3
elif opt.dataset == "cifar10":
    dataset = dset.CIFAR10(
        root=opt.data_folder,
        download=True,
        transform=transforms.Compose(
            [
                transforms.Resize(opt.image_size),
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        ),
    )
    nc = 3

elif opt.dataset == "mnist":
    dataset = dset.MNIST(
        root=opt.data_folder,
        download=True,
        transform=transforms.Compose(
            [
                transforms.Resize(opt.image_size),
                transforms.ToTensor(),
                transforms.Normalize((0.5,), (0.5,)),
            ]
        ),
    )
    nc = 1

elif opt.dataset == "fake":
    dataset = dset.FakeData(
        image_size=(3, opt.image_size, opt.image_size), transform=transforms.ToTensor()
    )
    nc = 3

assert dataset
dataloader = torch.utils.data.DataLoader(
    dataset, batch_size=opt.batch_size, shuffle=True, num_data_workers=int(opt.data_workers)
)

device = torch.device("cuda:0" if opt.cuda else "cpu")
n_gpu = int(opt.n_gpu)
nz = int(opt.nz)
ngf = int(opt.ngf)
ndf = int(opt.ndf)


# custom weights initialization called on net_generator and net_discriminator
def weights_init(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        torch.nn.init.normal_(m.weight, 0.0, 0.02)
    elif classname.find("BatchNorm") != -1:
        torch.nn.init.normal_(m.weight, 1.0, 0.02)
        torch.nn.init.zeros_(m.bias)


class Generator(nn.Module):
    def __init__(self, n_gpu):
        super(Generator, self).__init__()
        self.n_gpu = n_gpu
        self.main = nn.Sequential(
            # input is Z, going into a convolution
            nn.ConvTranspose2d(nz, ngf * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(ngf * 8),
            nn.ReLU(True),
            # state size. (ngf*8) x 4 x 4
            nn.ConvTranspose2d(ngf * 8, ngf * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 4),
            nn.ReLU(True),
            # state size. (ngf*4) x 8 x 8
            nn.ConvTranspose2d(ngf * 4, ngf * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 2),
            nn.ReLU(True),
            # state size. (ngf*2) x 16 x 16
            nn.ConvTranspose2d(ngf * 2, ngf, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf),
            nn.ReLU(True),
            # state size. (ngf) x 32 x 32
            nn.ConvTranspose2d(ngf, nc, 4, 2, 1, bias=False),
            nn.Tanh()
            # state size. (nc) x 64 x 64
        )

    def forward(self, input):
        if input.is_cuda and self.n_gpu > 1:
            output = nn.parallel.data_parallel(self.main, input, range(self.n_gpu))
        else:
            output = self.main(input)
        return output


net_generator = Generator(n_gpu).to(device)
net_generator.apply(weights_init)
if opt.net_generator != "":
    net_generator.load_state_dict(torch.load(opt.net_generator))
print(net_generator)


class Discriminator(nn.Module):
    def __init__(self, n_gpu):
        super(Discriminator, self).__init__()
        self.n_gpu = n_gpu
        self.main = nn.Sequential(
            # input is (nc) x 64 x 64
            nn.Conv2d(nc, ndf, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf) x 32 x 32
            nn.Conv2d(ndf, ndf * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*2) x 16 x 16
            nn.Conv2d(ndf * 2, ndf * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*4) x 8 x 8
            nn.Conv2d(ndf * 4, ndf * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*8) x 4 x 4
            nn.Conv2d(ndf * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, input):
        if input.is_cuda and self.n_gpu > 1:
            output = nn.parallel.data_parallel(self.main, input, range(self.n_gpu))
        else:
            output = self.main(input)

        return output.view(-1, 1).squeeze(1)


net_discriminator = Discriminator(n_gpu).to(device)
net_discriminator.apply(weights_init)
if opt.net_discriminator != "":
    net_discriminator.load_state_dict(torch.load(opt.net_discriminator))
print(net_discriminator)

criterion = nn.BCELoss()

fixed_noise = torch.randn(opt.batch_size, nz, 1, 1, device=device)
real_label = 1
fake_label = 0

# setup optimizer
optim_discriminator = optim.Adam(net_discriminator.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))
optim_generator = optim.Adam(net_generator.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))

if opt.dry_run:
    opt.n_iter = 1

for epoch in range(opt.n_iter):
    for i, data in enumerate(dataloader, 0):
        ############################
        # (1) Update D network: maximize log(D(x)) + log(1 - D(G(z)))
        ###########################
        # train with real
        net_discriminator.zero_grad()
        real_cpu = data[0].to(device)
        batch_size = real_cpu.size(0)
        label = torch.full(
            (batch_size,), real_label, dtype=real_cpu.dtype, device=device
        )
        output = net_discriminator(real_cpu)
        err_d_real = criterion(output, label)
        err_d_real.backward()
        d_x = output.mean().item()

        # train with fake
        noise = torch.randn(batch_size, nz, 1, 1, device=device)
        fake = net_generator(noise)
        label.fill_(fake_label)
        output = net_discriminator(fake.detach())
        err_d_fake = criterion(output, label)
        err_d_fake.backward()
        d_g_z1 = output.mean().item()
        err_discrimantor = err_d_real + err_d_fake
        optim_discriminator.step()

        ############################
        # (2) Update G network: maximize log(D(G(z)))
        ###########################
        net_generator.zero_grad()
        label.fill_(real_label)  # fake labels are real for generator cost
        output = net_discriminator(fake)
        err_generator = criterion(output, label)
        err_generator.backward()
        d_g_z2 = output.mean().item()
        optim_generator.step()

        print(
            "[%d/%d][%d/%d] Loss_D: %.4f Loss_G: %.4f D(x): %.4f D(G(z)): %.4f / %.4f"
            % (
                epoch,
                opt.n_iter,
                i,
                len(dataloader),
                err_discrimantor.item(),
                err_generator.item(),
                d_x,
                d_g_z1,
                d_g_z2,
            )
        )
        if i % 100 == 0:
            vutils.save_image(real_cpu, f"{opt.out_folder}/real_samples.png", normalize=True)
            fake = net_generator(fixed_noise)
            vutils.save_image(
                fake.detach(),
                "%s/fake_samples_epoch_%03d.png" % (opt.out_folder, epoch),
                normalize=True,
            )

        if opt.dry_run:
            break
    # do checkpointing
    torch.save(net_generator.state_dict(), "%s/net_generator_epoch_%d.pth" % (opt.out_folder, epoch))
    torch.save(net_discriminator.state_dict(), "%s/net_discriminator_epoch_%d.pth" % (opt.out_folder, epoch))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/store/store_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# third party
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2

# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/store/store_object.proto",
    package="syft.core.store",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n#proto/core/store/store_object.proto\x12\x0fsyft.core.store\x1a%proto/core/common/common_object.proto\x1a\x19google/protobuf/any.proto"\x82\x01\n\x12StoreGenericObject\x12\x38\n\x0fstorable_object\x18\x01 \x01(\x0b\x32\x1f.syft.core.store.StorableObject\x12\x32\n\x0cstore_object\x18\x02 \x01(\x0b\x32\x1c.syft.core.store.StoreObject"\xdf\x01\n\x0eStorableObject\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12\x1a\n\x12schematic_qualname\x18\x03 \x01(\t\x12"\n\x04\x64\x61ta\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x18\n\x10read_permissions\x18\x07 \x01(\x0c\x12\x1a\n\x12search_permissions\x18\x08 \x01(\x0c"r\n\x0bStoreObject\x12-\n\tcontainer\x18\x01 \x01(\x0b\x32\x1a.syft.core.store.Container\x12\x34\n\rsearch_engine\x18\x02 \x01(\x0b\x32\x1d.syft.core.store.SearchEngine"\xb1\x01\n\tContainer\x12\x36\n\x0e\x64ict_container\x18\x01 \x01(\x0b\x32\x1e.syft.core.store.DictContainer\x12\x34\n\rsql_container\x18\x02 \x01(\x0b\x32\x1d.syft.core.store.SqlContainer\x12\x36\n\x0egrid_container\x18\x03 \x01(\x0b\x32\x1e.syft.core.store.GridContainer"\x0e\n\x0cSearchEngine"\xa0\x01\n\rDictContainer\x12\x45\n\x0c\x64ict_mapping\x18\x01 \x03(\x0b\x32/.syft.core.store.DictContainer.DictMappingEntry\x1aH\n\x10\x44ictMappingEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x38\x01"\x0e\n\x0cSqlContainer"\x0f\n\rGridContainerb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_any__pb2.DESCRIPTOR,
    ],
)


_STOREGENERICOBJECT = _descriptor.Descriptor(
    name="StoreGenericObject",
    full_name="syft.core.store.StoreGenericObject",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="storable_object",
            full_name="syft.core.store.StoreGenericObject.storable_object",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="store_object",
            full_name="syft.core.store.StoreGenericObject.store_object",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=123,
    serialized_end=253,
)


_STORABLEOBJECT = _descriptor.Descriptor(
    name="StorableObject",
    full_name="syft.core.store.StorableObject",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="syft.core.store.StorableObject.id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data_type",
            full_name="syft.core.store.StorableObject.data_type",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="schematic_qualname",
            full_name="syft.core.store.StorableObject.schematic_qualname",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="syft.core.store.StorableObject.data",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="description",
            full_name="syft.core.store.StorableObject.description",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="tags",
            full_name="syft.core.store.StorableObject.tags",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="read_permissions",
            full_name="syft.core.store.StorableObject.read_permissions",
            index=6,
            number=7,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="search_permissions",
            full_name="syft.core.store.StorableObject.search_permissions",
            index=7,
            number=8,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=256,
    serialized_end=479,
)


_STOREOBJECT = _descriptor.Descriptor(
    name="StoreObject",
    full_name="syft.core.store.StoreObject",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="container",
            full_name="syft.core.store.StoreObject.container",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="search_engine",
            full_name="syft.core.store.StoreObject.search_engine",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=481,
    serialized_end=595,
)


_CONTAINER = _descriptor.Descriptor(
    name="Container",
    full_name="syft.core.store.Container",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="dict_container",
            full_name="syft.core.store.Container.dict_container",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sql_container",
            full_name="syft.core.store.Container.sql_container",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="grid_container",
            full_name="syft.core.store.Container.grid_container",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=598,
    serialized_end=775,
)


_SEARCHENGINE = _descriptor.Descriptor(
    name="SearchEngine",
    full_name="syft.core.store.SearchEngine",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=777,
    serialized_end=791,
)


_DICTCONTAINER_DICTMAPPINGENTRY = _descriptor.Descriptor(
    name="DictMappingEntry",
    full_name="syft.core.store.DictContainer.DictMappingEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="syft.core.store.DictContainer.DictMappingEntry.key",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="syft.core.store.DictContainer.DictMappingEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=882,
    serialized_end=954,
)

_DICTCONTAINER = _descriptor.Descriptor(
    name="DictContainer",
    full_name="syft.core.store.DictContainer",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="dict_mapping",
            full_name="syft.core.store.DictContainer.dict_mapping",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _DICTCONTAINER_DICTMAPPINGENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=794,
    serialized_end=954,
)


_SQLCONTAINER = _descriptor.Descriptor(
    name="SqlContainer",
    full_name="syft.core.store.SqlContainer",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=956,
    serialized_end=970,
)


_GRIDCONTAINER = _descriptor.Descriptor(
    name="GridContainer",
    full_name="syft.core.store.GridContainer",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=972,
    serialized_end=987,
)

_STOREGENERICOBJECT.fields_by_name["storable_object"].message_type = _STORABLEOBJECT
_STOREGENERICOBJECT.fields_by_name["store_object"].message_type = _STOREOBJECT
_STORABLEOBJECT.fields_by_name[
    "id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_STORABLEOBJECT.fields_by_name[
    "data"
].message_type = google_dot_protobuf_dot_any__pb2._ANY
_STOREOBJECT.fields_by_name["container"].message_type = _CONTAINER
_STOREOBJECT.fields_by_name["search_engine"].message_type = _SEARCHENGINE
_CONTAINER.fields_by_name["dict_container"].message_type = _DICTCONTAINER
_CONTAINER.fields_by_name["sql_container"].message_type = _SQLCONTAINER
_CONTAINER.fields_by_name["grid_container"].message_type = _GRIDCONTAINER
_DICTCONTAINER_DICTMAPPINGENTRY.fields_by_name[
    "value"
].message_type = google_dot_protobuf_dot_any__pb2._ANY
_DICTCONTAINER_DICTMAPPINGENTRY.containing_type = _DICTCONTAINER
_DICTCONTAINER.fields_by_name[
    "dict_mapping"
].message_type = _DICTCONTAINER_DICTMAPPINGENTRY
DESCRIPTOR.message_types_by_name["StoreGenericObject"] = _STOREGENERICOBJECT
DESCRIPTOR.message_types_by_name["StorableObject"] = _STORABLEOBJECT
DESCRIPTOR.message_types_by_name["StoreObject"] = _STOREOBJECT
DESCRIPTOR.message_types_by_name["Container"] = _CONTAINER
DESCRIPTOR.message_types_by_name["SearchEngine"] = _SEARCHENGINE
DESCRIPTOR.message_types_by_name["DictContainer"] = _DICTCONTAINER
DESCRIPTOR.message_types_by_name["SqlContainer"] = _SQLCONTAINER
DESCRIPTOR.message_types_by_name["GridContainer"] = _GRIDCONTAINER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StoreGenericObject = _reflection.GeneratedProtocolMessageType(
    "StoreGenericObject",
    (_message.Message,),
    {
        "DESCRIPTOR": _STOREGENERICOBJECT,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.StoreGenericObject)
    },
)
_sym_db.RegisterMessage(StoreGenericObject)

StorableObject = _reflection.GeneratedProtocolMessageType(
    "StorableObject",
    (_message.Message,),
    {
        "DESCRIPTOR": _STORABLEOBJECT,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.StorableObject)
    },
)
_sym_db.RegisterMessage(StorableObject)

StoreObject = _reflection.GeneratedProtocolMessageType(
    "StoreObject",
    (_message.Message,),
    {
        "DESCRIPTOR": _STOREOBJECT,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.StoreObject)
    },
)
_sym_db.RegisterMessage(StoreObject)

Container = _reflection.GeneratedProtocolMessageType(
    "Container",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONTAINER,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.Container)
    },
)
_sym_db.RegisterMessage(Container)

SearchEngine = _reflection.GeneratedProtocolMessageType(
    "SearchEngine",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHENGINE,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.SearchEngine)
    },
)
_sym_db.RegisterMessage(SearchEngine)

DictContainer = _reflection.GeneratedProtocolMessageType(
    "DictContainer",
    (_message.Message,),
    {
        "DictMappingEntry": _reflection.GeneratedProtocolMessageType(
            "DictMappingEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _DICTCONTAINER_DICTMAPPINGENTRY,
                "__module__": "proto.core.store.store_object_pb2"
                # @@protoc_insertion_point(class_scope:syft.core.store.DictContainer.DictMappingEntry)
            },
        ),
        "DESCRIPTOR": _DICTCONTAINER,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.DictContainer)
    },
)
_sym_db.RegisterMessage(DictContainer)
_sym_db.RegisterMessage(DictContainer.DictMappingEntry)

SqlContainer = _reflection.GeneratedProtocolMessageType(
    "SqlContainer",
    (_message.Message,),
    {
        "DESCRIPTOR": _SQLCONTAINER,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.SqlContainer)
    },
)
_sym_db.RegisterMessage(SqlContainer)

GridContainer = _reflection.GeneratedProtocolMessageType(
    "GridContainer",
    (_message.Message,),
    {
        "DESCRIPTOR": _GRIDCONTAINER,
        "__module__": "proto.core.store.store_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.GridContainer)
    },
)
_sym_db.RegisterMessage(GridContainer)


_DICTCONTAINER_DICTMAPPINGENTRY._options = None
# @@protoc_insertion_point(module_scope)
